#include <time.h>
#include <sys/sysctl.h>
#include <unistd.h>
#include <signal.h>
#include "general.h"




/*************/
/* FUNCTIONS */
/*************/

/* Append the list of labels to the sens-array of the current resource 
 * Dynamically resizes this array if necessary  */
void PS_appendToSens(int curr, const int* labels, int num_of_labels, Marks *marks) {
        marks[curr].sens.insert(marks[curr].sens.end(), &labels[0], &labels[num_of_labels]);
}
/* Append the list of labels to the imm-array of the current resource 
 * Dynamically resizes this array if necessary  */
void PS_appendToImm(int curr, const int* labels, int num_of_labels, Marks *marks) {
        marks[curr].imm.insert(marks[curr].imm.end(), &labels[0], &labels[num_of_labels]);
}


/* Append the list of labels leading to 2marked neighbors to the imm-array of the current resource 
Dynamically resizes this array if necessary  */
void PS_appendToImm_2marks(int curr, Graph *g, Marks *marks) {
	int i;
        int num_of_edges = g->n;
	for (i=0;i<num_of_edges;i++) {
		if (marks[g->neighbor(i)].marking == 2) {
			PS_appendToImm(curr, g->labels(i), g->labelCount(i), marks);
		}
	}
}


/*****************/
/* THE ALGORITHM */
/*****************/


void PS_step1 (Graph *g, int curr, int curr_index, int parent, Marks *marks, const Params *params) {
	int i;

	if (marks[curr].marking == 1 || marks[curr].marking == 3) {
		if (parent != -1)
			PS_appendToSens(parent, g[parent].labels(curr_index), g[parent].labelCount(curr_index), marks);
	}
	else if (marks[curr].marking == 2 || g[curr].n == 0) {
		marks[curr].marking = 2;
		if (parent != -1)
			PS_appendToImm(parent, g[parent].labels(curr_index), g[parent].labelCount(curr_index), marks);
	}
	else {
		marks[curr].marking = 1;
		for (i=0; i < g[curr].n; i++)
			PS_step1(g, g[curr].neighbor(i), i, curr, marks,params);
		if (marks[curr].subset()) {
			marks[curr].marking = 2;
			if (parent != -1)
				PS_appendToImm(parent, g[parent].labels(curr_index), g[parent].labelCount(curr_index), marks);
		}
		else {
			marks[curr].marking = 3;
			if (parent != -1)
				PS_appendToSens(parent, g[parent].labels(curr_index), g[parent].labelCount(curr_index), marks);
		}
	}
}

void PS_step2 (Graph *g, int curr, Marks *marks, const Params* params) {
	int j;

	if (marks[curr].marking == 3) {
		PS_appendToImm_2marks(curr, &g[curr], marks);
		if (marks[curr].subset()) {
			marks[curr].marking = 2;
			for (j=0;j<g[curr].parents_elements;j++) {
				if (marks[g[curr].parents[j]].marking == 3)
					PS_step2(g, g[curr].parents[j], marks, params);
			}
		}
	}
}

void clear_marks(Marks *marks, int n) {
	int i;
        Marks m;
	for (i=0;i<n;i++) {
		marks[i] = m; 
	}
}



int PS_run_algorithm_on_faulty_configuration(ResourceList* faulty_resources, const Params* params) {
    std::cout << "Starting algorithm for configuration with faulty channels: " << *faulty_resources << std::endl;
	// Declare variables
	int num_of_resources; 
	int i,j,dl;


	// Initialize assisting parameter, determine number of resources, reserve space for data structures
	num_of_resources = bitsize_of_resources(params);
	Graph *g = new Graph[num_of_resources];
	Marks *marks = new Marks[num_of_resources];
	int *nodeset = new int[num_of_resources];

	// Initialize graph and markings data structures
	zero_list(nodeset,num_of_resources);
    dl = fill_graph(g,nodeset,faulty_resources,params);
	if (dl == RET_DISC_ROUTING)
		goto TERMINATE_RUN_ALGORITHM;
	compute_parents(g, nodeset, num_of_resources);
	clear_marks(marks, num_of_resources);


	// Step 1
	for (i=0;i<num_of_resources;i++)
		if (nodeset[i] == 1)
    		PS_step1(g, i, -1, -1, marks, params);
	// Step 2
	for (i=0;i<num_of_resources;i++)
		if (nodeset[i] == 1)
			PS_step2(g,i,marks, params);

	// Find a 3-marked resource 
	dl = RET_DLF;
	for (i=0;i<num_of_resources;i++) {
		if (marks[i].marking == 3) {
			if (dl == RET_DLF) {
				dl = RET_DL_NOT_MIN;
				printf("Found deadlock under fault model:\n");
				int j= 0;
				while (faulty_resources->at(j).valid()) {
 					std::cout << faulty_resources->at(j++) << ", ";
				}
				std::cout << std::endl;
			}
 			
			std::cout << "Resource " << inst_convert_to_resource(i,params) << " can be in deadlock if it contains a packet with a destination in {";
			int printed=0;
			for (j=0;j<marks[i].sens.size()&&printed<10;j++) {
				int jj,found=0;
				for (jj=0;jj<marks[i].imm.size()&&found==0;jj++)
					found = (marks[i].imm[jj] == marks[i].sens[j] ? 1 : 0);
				if (found == 0) {
					inst_print_procnode (marks[i].sens[j], params);
					printed++;
          			std::cout << ",";
				}
			}
			if (printed == 10)
				printf("...");
			std::cout << "}" << std::endl;
		}
		else if (marks[i].marking != 2 && nodeset[i] == 1) {
			std::cout << "ERROR: resource " << inst_convert_to_resource(i,params) << " is marked " << marks[i].marking << ":" << std::endl;
		}
	}

	if (dl == RET_DL_NOT_MIN) {
		if (output)
			printf("Possible deadlock is actually reachable (but not necessarily minimal), by filling each channel with a header flit.\n");
	}
	else if (dl == RET_DLF) {
		if (output)
			std::cout << "The network is deadlock-free under configuration with faulty channels: " << *faulty_resources << std::endl;
	}
	else if (dl == RET_DISC_ROUTING) {
		if (output)
			printf("The network has a disconnected routing function.\n");
	}
	else
		assert(false);

TERMINATE_RUN_ALGORITHM:
	//write(fildes, &dl, sizeof(dl));

	for (i=0;i<num_of_resources;i++) {
		if (nodeset[i] != 0) {
			for (j=0;j<g[i].n;j++)
				free(g[i].labels(j));
			free(g[i].parents);
		}
	}
	delete[] g;
	delete[] nodeset;
	delete[] marks;
if (dl != RET_DLF && !ASSEMBLE_STATISTICS) exit(0);
    return dl;
}



