#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/wait.h>
#include <math.h>
#include <time.h>
#include <signal.h>

#include "general.h"

int main(int argc, char **argv) {
	while(argc--) {
		if (!strcmp(*argv, "min_dl_not_nec")) {
			MIN_DL_NOT_NEC = true;
			printf("Not finding minimum deadlock.\n");
		}
		else if (!strcmp(*argv, "verbose")) {
			VERBOSE = true;
			printf("Printing verbose output.\n");
		}
		else if (!strcmp(*argv, "do_not_check_reachability_of_possible_dl")) {
			DO_NOT_CHK_R_OF_POS_DL = true;
			printf("If a possible WHS specific deadlock is found, it is not checked whether this deadlock is reachable.\n");
		}
		else if (!strcmp(*argv, "report180")) {
			REPORT180 = true;
			printf("Checking for 180 turns. Report them and stop as soon as one is found.\n");
		}
		else if (!strcmp(*argv, "smt")) {
			SMT = true;
			printf("Use SMT solver instead of SAT solver. Works only in combination with the option `min_dl_not_nec'.\n");
		}
		else if (!strcmp(*argv, "local")) {
			LOCAL_MINIMA = true;
			printf("Reducing non-minimal deadlocks to local minima.\n");
		}
		else if (!strcmp(*argv, "break_at_error")) {
			ASSEMBLE_STATISTICS = false;
			printf("Exiting on first occurence of error/deadlock, instead of continueing and assembling quantative information.\n");
		}
		else if (!strcmp(*argv, "minimize_resources")) {
			MINIMIZE_RESOURCES = true;
			printf("Instead of minimizing the number of worms, minimize the number of resources occuring in the deadlock.\n");
		}
		else if (!strcmp(*argv, "PS")) {
			PS = true;
			printf("Looking for Packet Switching deadlocks, not for Wormhole Switching deadlocks. Incompatible with options `min_dl_not_nec', `do_not_check_reachability_of_possible_dl', `smt', `local', `break_at_error' or `minimize_resources'.\n");
		}
		else if (!strcmp(*argv, "help") || !strcmp(*argv, "?") || !strcmp(*argv, "h")) {
			printf("Executes algorithm to decide deadlock freedom of wormhole or packet networks.\n\nCommand line options:\nmin_dl_not_nec\n\tIf a possible wormhole deadlock is found, do determine whether this deadlock is actually reachable, but do not find the minimum deadlock.\nverbose\n\tPrint verbose output.\nsmt\n\tUse SMT solver instead of SAT solver. Works only in combination with the option `min_dl_not_nec'.\nlocal\n\tReducing non-minimal deadlocks to local minima. Works only in combination with the option `min_dl_not_nec'.\ndo_not_check_reachability_of_possible_dl\n\tIf a possible wormhole deadlock is found, do not check whether the deadlock is actually reachable.\nreport180\n\tCheck for 180 turns and stop the algorithm if one is found.\break_at_error\n\tExit on first occurence of error/deadlock, instead of continueing and assembling quantative information.\nminimize_resources\n\tInstead of minimizing the number of worms, minimize the number of resources occuring in the deadlock.\nPS\n\tLook for Packet Switching deadlocks, not for Wormhole Switching deadlocks. Incompatible with options `min_dl_not_nec', `do_not_check_reachability_of_possible_dl', `smt', `local', `break_at_error' or `minimize_resources'.\n");
			puts("");
			exit(0);
		}
		argv++;
	}


    // The starting time, for bookkeeping
    time_t start;
    time(&start);
    // Instantiate the topology parameters, i.e., the size of a mesh
    int i = 0;
    long long curr_config = 0;
    Params *params = inst_params();
    // Compute the number of different configurations to test
    long long num_configs = get_num_of_faulty_configurations(params);
    // Compute the number of cores
    int num_cores = num_of_cores();
    long long batch_size = ceil((double) num_configs/num_cores);
    fprintf(stderr, "num of cores = %i, batch_size = %lli -> total = %lli\n", num_cores, batch_size, num_configs);

/*
The following code is usefull for debugging:
Resource r = Resource(6,0,L);
Procnode d = Procnode(4,4);
//ResourceListN<2> fr;
//fr[0] = Resource(2,2,WIRELESS2);
ResourceList* neighbors = inst_routing(r,&d,params,NULL);//&fr);
int ii = 0;
while (neighbors->at(ii).valid()) {
    std::cout << neighbors->at(ii++) << "\n";
}
exit(0);
*/


	struct RET_table ret_t1 = {0,0,0,false};

    // We create a child process which creates `num_cores' batch processes, i.e., one for each core.
    // Each batch process calls the algorithm on a batch of configurations to test.

    // We create file descriptors for communication to the main process
    // The main process receives 0's and 1's from the batch processes through pfildes[0].
    // The batch processess write through pfildes[1].
    // A 0 means "no deadlock found: proceed".
    // A 1 means "deadlock found: terminate everything"
    int pfildes[2];
    pipe(pfildes);
    pid_t pID = fork();
    if (pID == 0) {
        // We're in a child process, we create `num_cores' batch processes and exit.
        close(pfildes[0]);

        for (curr_config=0; curr_config<num_configs; curr_config+=batch_size) {
            pid_t algo_pID = fork();
            if (algo_pID == 0) {
                struct RET_table ret_t2 = {0,0,0,false};
				run_algorithm_on_batch_of_faulty_configurations(curr_config,min(num_configs,curr_config+batch_size), /*pfildes[1],*/ &ret_t2, params);
                write(pfildes[1], &ret_t2, sizeof(ret_t2));
                fprintf(stderr, "CORE%i: Running time is %f secs.\n", (int) (curr_config/batch_size), clock()/(float)CLOCKS_PER_SEC);
				if (!ASSEMBLE_STATISTICS)
					kill(0,SIGTERM);//Hard kill of everything, this should be done cleaner!
                _exit(0);
            }
        }
        close(pfildes[1]);
        _exit(0);
    }
    else if (pID > 0) {
        // We're in the main process, reading 0's and 1's until either all batch processes have terminated
        // or a deadlock has been found.
        close(pfildes[1]);

        struct RET_table ret_t2;
        while (!ret_t1.terminate && read(pfildes[0], &ret_t2, sizeof(ret_t2)) > 0) {
			if (ASSEMBLE_STATISTICS) {
				add_ret_tables(&ret_t1, ret_t2);
			}
			else {
            	if (ret_t2.dl != 0 || ret_t2.disc_routing != 0 || ret_t2.terminate) {
					ret_t1.terminate = true;
					kill(pID,SIGKILL);
				}
			}
        }
        close(pfildes[0]);
    }
    else {
        fprintf(stderr, "ERROR: could not fork!\n");
    }

	if (ASSEMBLE_STATISTICS) {
		printf("Total number of configurations: %lli\n", num_configs);
		printf("Faulty configurations with deadlocks = %i\nFaulty configurations that are deadlock-free = %i\nFaulty configurations with disconnected routing = %i\n", ret_t1.dl, ret_t1.dlf, ret_t1.disc_routing);
	}


    time_t end;
    time(&end);
    fprintf(stderr, "\n\n");
    fprintf(stderr, "Starting time: %s", ctime(&start));
    fprintf(stderr, "End time: %s", ctime(&end));
    fprintf(stderr, "Elapsed time: %lli seconds\n", (long long int)(end-start));

    return 0;
}
