/**************/
/* ROUTING    */
/**************/

/*
A routing function which is deadlock-free and tolerant to two faulty channels.
*/

#include "abdulaziz_definitions.h"


/*
The routing function. Routes from channel c == (x,y,p) to processing node n == (dx,dy)
Returns a set of resources. As routing is adaptive, multiple next hops are returned.
This list is closed with a NULL.

In this case the assumtion is that East traffic and West traffic are maintained, which means that
if input port is from S1 or N1 (or W) then the destination x address has to be > than x address in current node. similar
if the input is S2 or N2 or ( or E)  the destination has to be < tha x address in  current node

*/

int is_border_channel(int x, int y, port p, const Params* dim) {
	return (x == 0 && p == W) || (x == dim->x-1 && p == E) || (y == 0 && p == N1) || (y == 0 && p == N2) || (y == dim->y-1 && p == S1) || (y == dim->y-1 && p == S2);
}
int is_ok(int x, int y, port p, ResourceList* frs, const Params* dim) {
	return x >=0 && y >= 0 && x < dim->x && y < dim->y && !is_faulty_resource(Resource(x,y,p), frs) && !is_border_channel(x,y,p,dim);
}

ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<3>;
    Resource *nexthops = hops->routed;

	// The coordinates of the destination
	int dx = d->x, dy = d->y;
	// The coordinates of the processing node at the end of channel c
	Procnode end = get_end(c);
	int nx = end.x, ny = end.y;
	int i = 0;
       	
	if (dx != nx || dy != ny) {          
		if (dx > nx && dy > ny) {
			// SOUTH-EAST
			if (is_ok(nx,ny,E,frs,dim) && ((dx>nx+1 || (dx==nx+1 && (is_ok(nx+1,ny,S1,frs,dim) || is_ok(nx+1,ny,S2,frs,dim))))))
				//Request east channel:
				nexthops[i++] = Resource(nx, ny, E);
			if (is_ok(nx,ny,S1,frs,dim) && is_ok(nx,ny+1,E,frs,dim) && (!is_ok(nx,ny,E,frs,dim) || is_ok(nx,ny,S2,frs,dim)))//NEW: third case.
				//Request south1 channel:
				nexthops[i++] = Resource(nx, ny, S1);

			if (i==0) {
				if (is_ok(nx,ny,N1,frs,dim))
					//Request north1 channel:
					nexthops[i++] = Resource(nx, ny, N1);
				else if (is_ok(nx,ny,S1,frs,dim))
					//Request south1 channel:
					nexthops[i++] = Resource(nx, ny, S1);
				else if (is_ok(nx,ny,S2,frs,dim))
					//Request south2 channel:
					nexthops[i++] = Resource(nx, ny, S2);
			}
		} 
		else if (dx > nx && dy < ny) {
			// NORTH-EAST
			if (is_ok(nx,ny,E,frs,dim) && ((dx>nx+1 || (dx==nx+1 && (is_ok(nx+1,ny,N1,frs,dim) || is_ok(nx+1,ny,N2,frs,dim))))))
				//Request east channel:
				nexthops[i++] = Resource(nx, ny, E);
			if (is_ok(nx,ny,N1,frs,dim) && is_ok(nx,ny-1,E,frs,dim) && (!is_ok(nx,ny,E,frs,dim) || is_ok(nx,ny,N2,frs,dim)))//NEW: third case.
				//Request north1 channel:
				nexthops[i++] = Resource(nx, ny, N1);
			if (i==0) {
				if (is_ok(nx,ny,S1,frs,dim))
					//Request south1 channel:
					nexthops[i++] = Resource(nx, ny, S1);
				else if (is_ok(nx,ny,N1,frs,dim))
					//Request north1 channel:
					nexthops[i++] = Resource(nx, ny, N1);
				else if (is_ok(nx,ny,N2,frs,dim))
					//Request north2 channel:
					nexthops[i++] = Resource(nx, ny, N2);
			}
		} 
		else if (dx > nx && dy == ny) {
			// EAST
			if (is_ok(nx,ny,E,frs,dim))
				//Request east channel:
				nexthops[i++] = Resource(nx, ny, E);
			else {
				if (is_ok(nx,ny,N1,frs,dim) && (is_ok(nx,ny-1,E,frs,dim) || is_ok(nx,ny-1,N1,frs,dim)))
					//Request north1 channel:
					nexthops[i++] = Resource(nx, ny, N1);
				if (is_ok(nx,ny,S1,frs,dim) && (is_ok(nx,ny+1,E,frs,dim) || is_ok(nx,ny+1,S1,frs,dim)))
					//Request south1 channel:
					nexthops[i++] = Resource(nx, ny, S1);
				if (i==0) {
					if (is_ok(nx,ny,N2,frs,dim) && !is_ok(nx,ny,S2,frs,dim) && !is_ok(nx,ny,S1,frs,dim) && !is_ok(nx,ny,N1,frs,dim))
						//Request north2 channel:
						nexthops[i++] = Resource(nx, ny, N2);
					else if (is_ok(nx,ny,S2,frs,dim)) //(dy > 0 && !is_ok(nx,ny,N2,frs,dim) && !is_ok(nx,ny,N1,frs,dim) && !is_ok(nx,ny,S1,frs,dim)) 
						//Request south2 channel:
						nexthops[i++] = Resource(nx, ny, S2);
				}
			}
		}
		else if (dx < nx && dy > ny) {
			// SOUTH-WEST
			if (is_ok(nx,ny,W,frs,dim) && ((dx<nx-1 || (dx==nx-1 && (is_ok(nx-1,ny,S1,frs,dim) || is_ok(nx-1,ny,S2,frs,dim))))))
				//Request west channel:
				nexthops[i++] = Resource(nx, ny, W);
			if (is_ok(nx,ny,S2,frs,dim) && is_ok(nx,ny+1,W,frs,dim) && (!is_ok(nx,ny,W,frs,dim) || is_ok(nx,ny,S1,frs,dim)))//NEW: third case.
			//if (is_ok(nx,ny,S2,frs,dim) && is_ok(nx,ny+1,W,frs,dim))
				//Request south2 channel:
				nexthops[i++] = Resource(nx, ny, S2);
			if (i==0) {
				if (is_ok(nx,ny,N2,frs,dim))
					//Request north2 channel:
					nexthops[i++] = Resource(nx, ny, N2);
				else if (is_ok(nx,ny,S2,frs,dim))
					//Request south2 channel:
					nexthops[i++] = Resource(nx, ny, S2);
				else
					//Request south1 channel:
					nexthops[i++] = Resource(nx, ny, S1);
			}
		}
		else if (dx < nx && dy < ny) {
			// NORTH-WEST
			if (is_ok(nx,ny,W,frs,dim) && ((dx<nx-1 || (dx==nx-1 && (is_ok(nx-1,ny,N1,frs,dim) || is_ok(nx-1,ny,N2,frs,dim))))))
				//Request west channel:
				nexthops[i++] = Resource(nx, ny, W);
			if (is_ok(nx,ny,N2,frs,dim) && is_ok(nx,ny-1,W,frs,dim) && (!is_ok(nx,ny,W,frs,dim) || is_ok(nx,ny,N1,frs,dim)))//NEW: third case.
			//if (is_ok(nx,ny,N2,frs,dim) && is_ok(nx,ny-1,W,frs,dim))
				//Request north2 channel:
				nexthops[i++] = Resource(nx, ny, N2);
			if (i==0) {
				if (is_ok(nx,ny,S2,frs,dim))
					//Request south2 channel:
					nexthops[i++] = Resource(nx, ny, S2);
				else if (is_ok(nx,ny,N2,frs,dim))
					//Request north2 channel:
					nexthops[i++] = Resource(nx, ny, N2);
				else if (is_ok(nx,ny,N1,frs,dim))
					//Request north1 channel:
					nexthops[i++] = Resource(nx, ny, N1);
			}
		} 
		else if (dx < nx && dy == ny) {
			// WEST
			if (is_ok(nx,ny,W,frs,dim))
				//Request west channel:
				nexthops[i++] = Resource(nx, ny, W);
			else {
				if (is_ok(nx,ny,N2,frs,dim) && (is_ok(nx,ny-1,W,frs,dim) || is_ok(nx,ny-1,N2,frs,dim)))
					//Request north2 channel:
					nexthops[i++] = Resource(nx, ny, N2);
				if (is_ok(nx,ny,S2,frs,dim) && (is_ok(nx,ny+1,W,frs,dim) || is_ok(nx,ny+1,S2,frs,dim)))
					//Request south2 channel:
					nexthops[i++] = Resource(nx, ny, S2);
				if (i==0) {
					if (is_ok(nx,ny,N1,frs,dim) && !is_ok(nx,ny,S1,frs,dim) && !is_ok(nx,ny,S2,frs,dim) && !is_ok(nx,ny,N2,frs,dim))
						//Request north1 channel:
						nexthops[i++] = Resource(nx, ny, N1);
					else if (is_ok(nx,ny,S1,frs,dim)) //(ny > 0 && !is_ok(nx,ny,N1,frs,dim) && !is_ok(nx,ny,N2,frs,dim) && !is_ok(nx,ny,S2,frs,dim)) 
						//Request south1 channel:
						nexthops[i++] = Resource(nx, ny, S1);
				}
			}
		}
		else if (dx == nx && dy > ny) {
			// SOUTH
			nexthops[0] = Resource(nx, ny, S1);
			nexthops[1] = Resource(nx, ny, S2);
			if (!is_ok(nx,ny,S1,frs,dim) && !is_ok(nx,ny,S2,frs,dim)) {
				nexthops[1] = Resource();
				if (is_ok(nx,ny,E,frs,dim))
					nexthops[0] = Resource(nx, ny, E);
				else if (nx > 0)
					nexthops[0] = Resource(nx, ny, W);
			}
		}
		else if (dx == nx && dy < ny) {
			// NORTH
			nexthops[0] = Resource(nx, ny, N1);
			nexthops[1] = Resource(nx, ny, N2);
			if (!is_ok(nx,ny,N1,frs,dim) && !is_ok(nx,ny,N2,frs,dim)) {
				nexthops[1] = Resource();
				if (is_ok(nx,ny,E,frs,dim))
					nexthops[0] = Resource(nx, ny, E);
				else if (nx > 0)
					nexthops[0] = Resource(nx, ny, W);
			}
		}
	}
	return hops;
}

