/**************/
/* ROUTING    */
/**************/

/*
Routing function for Wireless NoC from UCI.
*/

#include "wireless_faults.c"


// This routing function defines Wireless Routing for a 10x10 NePA architecture.
// We choose a delta:
const int delta = 4;


// Returns true iff the port indicates a wireless channel
bool wireless (const char p) {
	return p == WIRELESS1 || p == WIRELESS2 || p == WIRELESS3 || p == WIRELESS4;
}
// Returns true if [nx,ny] is in the same quadrant as [dx,dy]
bool in_same_quadrant (int nx, int ny, int dx, int dy, const Params* dim) {
	return	(nx < (dim->x /2) && dx < (dim->x /2) && ny < (dim->y /2) && dy < (dim->y /2)) ||
			(nx < (dim->x /2) && dx < (dim->x /2) && ny >= (dim->y /2) && dy >= (dim->y /2)) ||
			(nx >= (dim->x /2) && dx >= (dim->x /2) && ny < (dim->y /2) && dy < (dim->y /2)) ||
			(nx >= (dim->x /2) && dx >= (dim->x /2) && ny >= (dim->y /2) && dy >= (dim->y /2));
}



// The default XY routing function: 
ResourceList* XY_routing (const Resource &c, int dx, int dy, const Params* dim, int subnetwork) {
    ResourceList *hops = new ResourceListN<4>;
    Resource *nexthops = hops->routed;
	int i = 0;

	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;

	switch (subnetwork) {
		case 1:
			if (dx > nx)
				nexthops[i++] = Resource(nx, ny, E);
			else if (dx < nx)
				nexthops[i++] = Resource(nx, ny, W);
			else if (dy < ny)
				nexthops[i++] = Resource(nx, ny, N1);
			else if (dy > ny)
				nexthops[i++] = Resource(nx, ny, S1);
			break;
		case 2:
			if (dx > nx)
				nexthops[i++] = Resource(nx, ny, E);
			else if (dx < nx)
				nexthops[i++] = Resource(nx, ny, W);
			else if (dy < ny)
				nexthops[i++] = Resource(nx, ny, N2);
			else if (dy > ny)
				nexthops[i++] = Resource(nx, ny, S2);
			break;
	}
	return hops;
}



// The wireless routing extension:
// Use NePA_routing to route towards the wireless router at the center of the quadrant.
// Then transmit the message wirelessly.
ResourceList* WNoC_routing (const Resource &c, int dx, int dy, const Params* dim) {
	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;

	// First determine the closest wireless router
	int wx = nx < (dim->x /2) ? 2 : 7;
	int wy = ny < (dim->y /2) ? 2 : 7;
	
	if (nx != wx || ny != wy)
		// If the packet has not arrived at the wireless router, route it towards it
		return XY_routing(c, wx, wy, dim, 1);
	else {
		// If the packet has arrived at the wireless router, transmit it wirelessly
		// Determine which wireless router:
		port p = dx < (dim->x /2) ? (dy < (dim->y /2) ? WIRELESS4 : WIRELESS3) : (dy < (dim->y /2) ? WIRELESS1 : WIRELESS2);
		// Return a next hop <dx,dy,p>
    	ResourceList *hops = new ResourceListN<2>;
		hops->routed[0] = Resource(nx,ny,p);
		return hops;
	}
}


// Compute the distance from [nx,ny] to [dx,dy] using wires.
int wired_travelling_distance(int nx, int ny, int dx, int dy, const Params* dim) {
	return abs(dx - nx) + abs(dy - ny);
}
// Compute the distance from [nx,ny] to [dx,dy] using wireless transmission.
int wireless_travelling_distance(int nx, int ny, int dx, int dy, const Params* dim) {
	// First determine the closest wireless router
	int wx = nx < (dim->x /2) ? 2 : 7;
	int wy = ny < (dim->y /2) ? 2 : 7;
	// Secondly, determine the receiving wireless router
	int wrecx = dx < (dim->x /2) ? 2 : 7;
	int wrecy = dy < (dim->y /2) ? 2 : 7;
	// Compute the distance to this wireless router
	int distance = abs(wx - nx) + abs(wy - ny);
	// Add one hop for the wireless transmission;
	distance++;
	// Compute the distance from the receiving wireless router to the destination.
	distance += abs(dx - wrecx) + abs(dy - wrecy);
	return distance;
}



// The final routing function:
ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
	// The coordinates of the destination
	int dx = d->x, dy = d->y;
	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;
	// The distances between source and destination
	int H_B = wired_travelling_distance(nx, ny, dx, dy, dim);
	int H_W = wireless_travelling_distance(nx, ny, dx, dy, dim);
       	
	if (dx != nx || dy != ny) {
		// If the wireless distance is significantly less than the wired distance, choose the wireless route
		// I also need a check that the destination is not in the current quadrant.
		if (H_W < H_B + delta && !in_same_quadrant(nx, ny, dx, dy, dim))
			return WNoC_routing(c, dx, dy, dim);
		else
			return XY_routing(c, dx, dy, dim, 2);
	}
	return new ResourceListN<1>;
}

